package cn.ischoolcode.javabook.dushichat;

/* 
 * 爱校码
 * http://www.ischoolcode.cn/
 *  
 */

import java.applet.*;
import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * 此类表示在客户端的用户连接。所有客户端的通信
 * 是通过这个类进行. 
 * 
 * @author zcj
 */
public class DushiClientConnection extends DushiConnection {
	private Class thisClass = DushiClientConnection.class;
	private DushiPanel mainPanel;
	private Vector ignoredUsers = new Vector();
	protected Vector userList = new Vector();
	protected boolean stop = false;

	public DushiClientConnection(Socket s, DushiPanel panel) throws IOException {
		super(null, "dushi 聊天客户端连接", s);

		mainPanel = panel;

		synchronized (iStream) {
			// 我们将把我们的首选协议版本发送到服务器。
			// 服务器将使用它希望我们使用的任何版本（
			// 永远不会超过我们要求的版本）进行响应
			sendProtocol(protocolVersion);

			if (iStream.readShort() == DushiCommand.SETPROTO)
				receiveProtocol();
			else {
				// 在断开/连接几次后似乎在这里丢失了它。这是一个变通办法。
				// 只需在不协商协议的情况下排出流即可。设置一个对话框
				// 让用户知道他们的客户端很糟糕。
				iStream.readFully(new byte[iStream.available()]);

				new DushiTextDialog(mainPanel.parentWindow,
						mainPanel.strings.get(thisClass, "warning"),
						mainPanel.strings.get(thisClass, "lostdata"), 60, 15,
						TextArea.SCROLLBARS_VERTICAL_ONLY, false,
						mainPanel.strings.get("dismiss"));
			}
		}

		// 开始监听服务器上的内容
		start();

		// 现在向服务器发送有关该用户的一些信息
		sendUserInfo();

		// 最后，启动看门狗来监控连接
		if (protocolVersion >= 2.2)
			watchDog = new WatchDog(this);
	}

	public void run() {
		while (online) {
			try {
				parseCommand();
			} catch (EOFException e) {
				// 这种情况偶尔会发生，就好像Java流在实际读取
				// 之前认为有什么东西可以立即读取一样……
				// 只需重置到标记的位置
				if (online) {
					try {
						iStream.reset();
					} catch (IOException f) {
						lostConnection();
					}
				}
				return;
			} catch (IOException e) {
				lostConnection();
			} catch (Exception e) {
				// 这里发现的任何其他异常都肯定是某种错误。
				// 请打印出来。
				e.printStackTrace();

				// 如果可能的话，尽量忽略它
			}

			yield();
		}
	}

	private void parseCommand() throws IOException {
		// 这个例程计算出要发送的命令，并将其发送到下面的相应子例程。

		short commandType = 0;
		boolean foundCommand = false;

		synchronized (iStream) {

			// 我们需要在这个“同步”块中进行所有输入处理，因为我们在
			// 这里读取命令的第一部分，然后在子例程中读取其余部分。

			//标记当前点，以便在必要时返回到该点。
			iStream.mark(Integer.MAX_VALUE);

			// 试着从流里读进一段。这表示命令类型。
			commandType = iStream.readShort();

			// 记录我们收到了什么。
			lastReceiveTime = System.currentTimeMillis();

			// 确保我们在阻止输入时没有断开连接
			if (!online)
				return;

			// 现在我们有了一个要读取的命令，我们将调用适当的
			// 客户端例程来解释它。

			if (!foundCommand && (protocolVersion >= 2.2)) {
				//协议版本2.2支持以下命令。有关协议定义，请参阅文件DushiCommand.java。

				// 命令SENDEMAIL从不从服务器发送到客户端（反之亦然），
				// 因此我们在这里不需要担心它们。

				foundCommand = true;

				switch (commandType) {

				  case DushiCommand.FILE: {
					// 我们收到启动文件传输的请求
					receiveFile();
					break;
				  }
				  case DushiCommand.FILEACCEPT: {
					// 我们收到关于用户是否已接受我们的文件传输的通知。
					receiveFileAccept();
					break;
				  }
				  case DushiCommand.FILEBLOCK: {
					// 我们收到一个文件块，它是文件传输的一部分。
					receiveFileBlock();
					break;
				  }
				  case DushiCommand.FILECANCEL: {
					// 我们收到一个文件传输已被取消的通知。
					receiveFileCancel();
					break;
				  }
				  case DushiCommand.IMAGEBLOCK: {
					// 我们正在接收一个图像块，它是图像传输的一部分。
					receiveImageBlock();
					break;
				  }
				  default: {
					foundCommand = false;
					break;
				  }
			    }

			}

			if (!foundCommand && (protocolVersion >= 2.0)) {
				// 协议版本2.0支持以下命令

				// LEAVEMESS、READMESS和EXIT命令从不从服务器发送到客户端
				//（反之亦然），所以我们在这里不需要担心它们。

				foundCommand = true;

				switch (commandType) {

				  case DushiCommand.NOOP: {
					// 不执行任何操作
					break;
				  }
				  case DushiCommand.PING: {
					// 服务器正在向我们发送ping。
					break;
				  }
				  case DushiCommand.CONNECT: {
					// 服务器告诉我们有一个新用户连接。
					receiveConnect();
					break;
				  }
				  case DushiCommand.USERINFO: {
					// 服务器正在发送有关用户（可能是我们）的信息。
					receiveUserInfo();
					break;
				  }
				  case DushiCommand.SERVERMESS: {
					// 服务器正在向我们发送“对话框”消息
					receiveServerMess();
					break;
				  }
				  case DushiCommand.DISCONNECT: {
					// 有人正在断开连接。也许是我们：）
					receiveDisconnect();
					break;
				  }
				  case DushiCommand.ROOMLIST: {
					// 服务器正在向我们提供当前聊天室的列表。
					receiveRoomList();
					break;
				  }
				  case DushiCommand.INVITE: {
					// 有人邀请我们加入聊天室
					receiveInvite();
					break;
				  }
				  case DushiCommand.ENTERROOM: {
					// 有人正在进入聊天室
					receiveEnterRoom();
					break;
				  }
				  case DushiCommand.BOOTUSER: {
					// 我们被赶出了聊天室
					receiveBootUser();
					break;
				  }
				  case DushiCommand.BANUSER: {
					// 我们已被禁止进入我们请求的聊天室
					receiveBanUser();
					break;
				  }
				  case DushiCommand.ALLOWUSER: {
					// 有人允许我们进入聊天室
					receiveAllowUser();
					break;
				  }
				  case DushiCommand.ACTIVITY: {
					// 有人在打字、画画等。
					receiveActivity();
					break;
				  }
				  case DushiCommand.CHATTEXT: {
					// 有来自另一个用户的传入聊天文本
					receiveChatText();
					break;
				  }
				  case DushiCommand.LINE: {
					// 有人在画布上画了一条线
					receiveLine();
					break;
				  }
				  case DushiCommand.RECT: {
					// 有人在画布上画了一个长方形
					receivePoly(DushiCommand.RECT);
					break;
				  }
				  case DushiCommand.OVAL: {
					// 有人在画布上画了一个椭圆形
					receivePoly(DushiCommand.OVAL);
					break;
				  }
				  case DushiCommand.DRAWTEXT: {
					// 有人在画布上画了一些文字
					receiveDrawText();
					break;
				  }
				  case DushiCommand.IMAGE: {
					// 有人在画布上贴了一张图片
					receiveImage();
					break;
				  }
				  case DushiCommand.CLEARCANV: {
					// 有人清理了画布
					receiveClearCanv();
					break;
				  }
				  case DushiCommand.PAGEUSER: {
					// 有人在传呼我们
					receivePageUser();
					break;
				  }
				  case DushiCommand.INSTANTMESS: {
					// 服务器正在向我们发送即时消息
					receiveInstantMess();
					break;
				  }
				  case DushiCommand.STOREDMESS: {
					// 服务器正在向我们发送一个消息
					receiveStoredMess();
					break;
				  }
				  case DushiCommand.ERROR: {
					// 有人拒绝了我们的页面
					receiveError();
					break;
				  }
				  default: {
					foundCommand = false;
					break;
				  }
				}
			}

			if (!foundCommand) {
				//哎呀。我们不理解这个命令
				byte[] foo = new byte[iStream.available()];
				iStream.readFully(foo);

				//由于我们不了解服务器发送给我们的内容，我们是否
				//试图连接到不会说我们语言的Dushi Chat v1.x服务器？
				String stringFoo = new String(foo);
				if (stringFoo.startsWith("lcome to Dushi")) {
					shutdown(false);
					mainPanel.offline();
					notifyObsoleteV1();
					return;
				}

				System.out.println(mainPanel.strings.get(thisClass,
						"unknowncommand") + " " + commandType);
			}
		}
	}

	/**
	 * 此功能用于通知客户端此服务器是Dushi Chat v1.x服务器，
	 * 无法接受来自此客户端的连接。
	 */
	private void notifyObsoleteV1() {
		new DushiOkDialog(mainPanel.parentWindow, mainPanel.strings.get(
				thisClass, "obsolete"), mainPanel.strings.get(thisClass,
				"pleaseupgrade"), mainPanel.strings.get("ok"), true)
				.setVisible(true);
	}

	/**
	 * 根据用户ID在列表中查找用户。
	 */
	private DushiUser findUser(int userId) {
		DushiUser tmpUser = null;
		DushiUser returnUser = null;

		// 是 "nobody"吗?
		if (userId == 0)
			return (null);

		// 浏览已连接用户的列表，查找具有正确id的用户。
		for (int count = 0; count < userList.size(); count++) {
			tmpUser = (DushiUser) userList.elementAt(count);

			if (tmpUser.id == userId) {
				returnUser = tmpUser;
				break;
			}
		}

		return (returnUser);
	}

	/**
	 * 根据用户的用户名在列表中查找用户。
	 */
	protected DushiUser findUser(String userName) {
		DushiUser tmpUser = null;
		DushiUser returnUser = null;

		// 浏览已连接用户的列表，查找具有正确id的用户。
		for (int count = 0; count < userList.size(); count++) {
			tmpUser = (DushiUser) userList.elementAt(count);

			if (tmpUser.name.equals(userName)) {
				returnUser = tmpUser;
				break;
			}
		}

		return (returnUser);
	}

	/**
	 * 这将从输入流中读取用户id并返回相应的用户对象。如果用户id为零（“nobody”）
	 * 或在列表中找不到用户，则返回null。
	 */
	private DushiUser readUser() throws IOException {
		int userId = 0;

		// 用户的用户ID是什么？
		userId = iStream.readInt();

		return (findUser(userId));
	}

	/**
	 * 忽略特定用户。
	 */
	protected void addToIgnored(String name) {
		// 用户是否已被忽略？
		if (!isIgnoredUser(name))
			ignoredUsers.addElement(name);

		// 输出消息
		mainPanel.messagesArea.append("<<"
				+ mainPanel.strings.get(thisClass, "ignoreusers") + " ");
		for (int count = 0; count < ignoredUsers.size(); count++) {
			mainPanel.messagesArea.append((String) ignoredUsers
					.elementAt(count));
			if (count < (ignoredUsers.size() - 1))
				mainPanel.messagesArea.append(", ");
		}
		mainPanel.messagesArea.append(">>\n");
	}

	/**
	 * 此例程将从用户列表中选择的用户添加到我们的忽略用户列表中。
	 */
	protected void addToIgnored() {
		String[] selectedUsers;
		int numberUsers = 0;

		// 获取所选用户名的列表。
		selectedUsers = mainPanel.sendToList.getSelectedItems();

		// 有多少？
		numberUsers = selectedUsers.length;

		// 循环
		for (int count1 = 0; count1 < numberUsers; count1++) {
			// 在我们的用户列表中查找此用户
			for (int count2 = 0; count2 < userList.size(); count2++) {
				DushiUser tmp = (DushiUser) userList.elementAt(count2);

				if (selectedUsers[count1].equals(tmp.name))
					addToIgnored(tmp.name);
			}
		}
	}

	/**
	 * 如果提供的用户名在我们的忽略用户列表中，则返回true。
	 */
	private boolean isIgnoredUser(String who) {
		for (int count = 0; count < ignoredUsers.size(); count++)
			if (((String) ignoredUsers.elementAt(count)).equals(who))
				return (true);
		return (false);
	}

	//以下例程将发送和接收各种命令类型

	/**
	 * 服务器正在告诉我们要使用哪种协议。
	 */
	private void receiveProtocol() throws IOException {
		protocolVersion = iStream.readDouble();
	}

	/**
	 * 一个新用户已连接。我们只使用它来输出一条消息，说明用户已经连接；
	 * 我们得到一个USERINFO命令，它稍后会告诉我们有关用户的信息。
	 */
	private void receiveConnect() throws IOException {
		String userName = iStream.readUTF();

		mainPanel.messagesArea.append("<<"
				+ mainPanel.strings.get(thisClass, "newuser") + " \""
				+ userName + "\" "
				+ mainPanel.strings.get(thisClass, "connected") + ">>\n");
	}

	/**
	 * 向服务器发送有关此用户的一些信息。
	 */
	private void sendUserInfo() {
		// 发送一个为0的ID，因为我们还没有被分配一个ID。
		super.sendUserInfo(0, mainPanel.name, mainPanel.encryptedPassword,
				mainPanel.passwordEncryptor.canEncrypt, mainPanel.email,
				mainPanel.additional, mainPanel.strings.language);
	}

	/**
	 * 服务器正在发送有关某个用户的新信息。
	 */
	private void receiveUserInfo() throws IOException {
		int tmpId = 0;
		String tmpName = "";
		String tmpEmail = "";
		String tmpAdditional = "";
		String tmpLanguage = "zh"; // 默认为中文
		DushiUser newUser;

		tmpId = iStream.readInt();
		tmpName = iStream.readUTF();
		// 密码字段将为空
		iStream.readUTF();
		iStream.readBoolean();

		if (protocolVersion >= 2.2)
			// 读取用户的电子邮件地址
			tmpEmail = iStream.readUTF();

		tmpAdditional = iStream.readUTF();

		if (protocolVersion >= 2.1)
			// 阅读用户的首选语言。
			tmpLanguage = iStream.readUTF();

		// 用户名是我们的吗？如果是这样，则服务器会告诉我们自己的用户id号。
		if (tmpName.equals(mainPanel.name)) {
			mainPanel.clientId = tmpId;
		} else {
			// 用户名是我们的吗？如果是这样，则服务器会告诉我们自己的用户id号。
			newUser = new DushiUser(tmpId, tmpName, "", tmpEmail,
					tmpAdditional, tmpLanguage);

			// 将用户添加到我们的用户集合
			userList.addElement(newUser);
		}
	}

	/**
	 * 服务器正在向我们发送消息。获取消息。
	 */
	private void receiveServerMess() throws IOException {
		String message = "";

		message = iStream.readUTF();

		// 制作一个包含消息的对话框

		new DushiOkDialog(mainPanel.parentWindow, mainPanel.strings.get(
				thisClass, "servermessage"), message,
				mainPanel.strings.get("ok"), false).setVisible(true);
	}

	/**
	 * 告诉服务器我们正在断开连接。
	 */
	protected void sendDisconnect() {
		super.sendDisconnect(mainPanel.clientId, "");
	}

	/**
	 * 服务器告诉我们有人（可能是我们）正在断开连接。
	 */
	private void receiveDisconnect() throws IOException {
		int tmpId = 0;
		DushiUser tmpUser;
		String disconnectMess = "";
		java.awt.List list = mainPanel.sendToList;

		tmpId = iStream.readInt();
		disconnectMess = iStream.readUTF();
		iStream.readFully(new byte[iStream.available()]);

		// 是谁？是我们吗？
		if ((tmpId == mainPanel.clientId) || (tmpId == 0)) {
			stop = true;
			shutdown(false);
			mainPanel.offline();

			// 如果是我们，制作一个带有断开连接消息的对话框。

			if (disconnectMess.equals(""))
				disconnectMess = mainPanel.strings.get(thisClass, "noreason");

			new DushiOkDialog(mainPanel.parentWindow, mainPanel.strings.get(
					thisClass, "disconnected"), disconnectMess,
					mainPanel.strings.get("ok"), true).setVisible(true);
		} else {
			// 其他一些用户已断开连接。输出此用户已离开聊天室的消息。

			tmpUser = findUser(tmpId);

			if (tmpUser == null)
				// 不知道这是谁。忽略。
				return;

			mainPanel.messagesArea.append("<<" + tmpUser.name + " "
					+ mainPanel.strings.get(thisClass, "isdisconnecting")
					+ ">>\n");

			// 从“当前发送到”列表中删除此名称
			synchronized (list) {
				for (int count2 = 0; count2 < list.getItemCount(); count2++) {
					if (list.getItem(count2).equals(tmpUser.name)) {
						if (list.isIndexSelected(count2))
							list.select(0);
						list.remove(count2);
						break;
					}
				}

				//如果列表中没有剩余内容，请确保选中“全部发送”复选框。
				if (list.getSelectedItems().length == 0)
					mainPanel.sendToAllCheckbox.setState(true);
			}

			// 将此名称从其聊天室信息的用户列表中删除。
			DushiRoomInfo roomInfo = null;

			if (mainPanel.roomInfoArray != null) {
				for (int count2 = 0; count2 < mainPanel.roomInfoArray.length; count2++)
					// 这是用户的聊天室吗？
					if (mainPanel.roomInfoArray[count2].name
							.equals(tmpUser.chatroomName)) {
						roomInfo = mainPanel.roomInfoArray[count2];
						break;
					}
				if (roomInfo != null) {
					// 从聊天室的用户列表中删除该用户。
					roomInfo.userNames.removeElement(tmpUser.name);
					roomInfo.userNames.trimToSize();
				}
			}

			// 同时从我们的用户列表中删除此名称。
			userList.removeElement((Object) tmpUser);
			userList.trimToSize();

			// 如果有一个DushiRoomControlDialog，告诉它更新自己。
			if (mainPanel.roomControlDialog != null)
				mainPanel.roomControlDialog.updateLists();
		}
	}

	/**
	 * 这将要求服务器向我们发送当前聊天室的列表。
	 */
	protected void sendRoomList() {
		super.sendRoomList(null);
	}

	/**
	 * 这将从服务器接收聊天室列表，将其放置在一个数组中
	 *（并将数组提供给聊天室对话框（如果存在））。
	 */
	private void receiveRoomList() throws IOException {
		int howManyRooms = 0;
		DushiRoomInfo[] roomList;
		DushiRoomInfo tmp = null;

		// 有多少聊天室？
		howManyRooms = iStream.readShort();

		// 既然我们知道了数字，我们就可以创建数组
		roomList = new DushiRoomInfo[howManyRooms];

		// 现在循环并填写每个DushiRoomInfo结构中的信息
		for (int count1 = 0; count1 < howManyRooms; count1++) {
			tmp = new DushiRoomInfo();

			tmp.name = iStream.readUTF();
			tmp.creatorName = iStream.readUTF();
			tmp.priv = iStream.readBoolean();
			tmp.invited = iStream.readBoolean();
			int numUsers = iStream.readInt();

			// 填写用户名数组
			for (int count2 = 0; count2 < numUsers; count2++)
				tmp.userNames.addElement(iStream.readUTF());

			roomList[count1] = tmp;

			if (tmp.name.equals(mainPanel.currentRoom.name))
				mainPanel.currentRoom = tmp;
		}

		// 保存聊天室信息列表
		mainPanel.roomInfoArray = roomList;

		// 现在通知聊天室对话框（如果存在），以便在那里显示更改。
		if (mainPanel.roomsDialog != null)
			mainPanel.roomsDialog.receivedList();

		// 如果有一个DushiRoomControlDialog，告诉它更新自己。
		if (mainPanel.roomControlDialog != null)
			mainPanel.roomControlDialog.updateLists();
	}

	/**
	 * 告诉服务器我们想邀请用户进入我们的聊天室。
	 */
	protected void sendInvite(int userId, String roomName) {
		super.sendInvite(mainPanel.clientId, roomName, userId);
	}

	/**
	 * 此用户已被邀请加入聊天室。
	 */
	private void receiveInvite() throws IOException {
		DushiUser fromUser = null;
		String roomName = null;

		// 是谁？
		fromUser = readUser();

		roomName = iStream.readUTF();
		iStream.readInt(); // 丢出我们的用户id

		if (fromUser == null)
			// 确认。没有这样的用户。如果有人在适当的时候注销，就会发生这种情况。
			return;

		if (isIgnoredUser(fromUser.name))
			return;

		// 制作一个包含消息的对话框，以便用户知道他们已被邀请加入。
		new DushiOkDialog(mainPanel.parentWindow, mainPanel.strings.get(
				thisClass, "invitation"), mainPanel.strings.get(thisClass,
				"invitedtoroom")
				+ " \""
				+ roomName
				+ "\" "
				+ mainPanel.strings.get(thisClass, "byuser")
				+ " "
				+ fromUser.name, mainPanel.strings.get("ok"), false)
				.setVisible(true);
	}

	/**
	 * 告诉服务器我们想加入不同的聊天室。
	 */
	protected void sendEnterRoom(String roomName, boolean priv, String password) {
		super.sendEnterRoom(mainPanel.clientId, roomName, priv, password,
				mainPanel.passwordEncryptor.canEncrypt);
	}

	/**
	 * 一些用户（可能是我们）刚刚进入聊天室。
	 */
	private void receiveEnterRoom() throws IOException {
		int userId = 0;
		String newRoomName = "";
		DushiUser tmpUser;
		java.awt.List list = mainPanel.sendToList;

		userId = iStream.readInt();
		newRoomName = iStream.readUTF();
		iStream.readBoolean();
		iStream.readUTF(); // 忽略空密码
		iStream.readBoolean(); // 未加密

		// 用户是谁？如果是我们，我们需要设置聊天室的名称。
		// 如果是另一个用户，我们需要设置他们聊天室的名称，
		// 如果他们和我们在同一个聊天室，则将他们添加到我们
		// 的“发送到”列表中。

		if (userId == mainPanel.clientId) {
			// 是我们。
			if (!mainPanel.currentRoom.name.equals(""))
				mainPanel.messagesArea.append("<<"
						+ mainPanel.strings.get(thisClass, "entering") + " \""
						+ newRoomName + "\">>\n");

			// 清除“发送到”列表。
			if (mainPanel.sendToList.getItemCount() > 0)
				mainPanel.sendToList.removeAll();
			mainPanel.sendToAllCheckbox.setState(true);

			// 清除画布。
			mainPanel.canvas.clear();

			// 还有谁在我们的新房间里？将它们添加到我们的“发送到”列表中。
			for (int count1 = 0; count1 < userList.size(); count1++) {
				tmpUser = (DushiUser) userList.elementAt(count1);

				if (tmpUser.chatroomName.equals(newRoomName))
					mainPanel.sendToList.add(tmpUser.name);
			}

			// 获取我们要进入的房间的房间信息。

			for (int count1 = 0; count1 < mainPanel.roomInfoArray.length; count1++) {
				DushiRoomInfo roomInfo = mainPanel.roomInfoArray[count1];

				if (roomInfo.name.equals(newRoomName)) {
					// 找到了
					mainPanel.currentRoom = roomInfo;
					break;
				}
			}

			// 如果我们打开了一个聊天室控制对话框，我们应该丢弃它。
			// 但是，如果我们也拥有新的聊天室，我们也应该加载另一个。
			if (mainPanel.roomControlDialog != null) {
				mainPanel.roomControlDialog.dispose();
				mainPanel.roomControlDialog = null;
			}
			if (mainPanel.currentRoom.creatorName.equals(mainPanel.name))
				mainPanel.roomControlDialog = new DushiRoomControlDialog(
						mainPanel);

			return;
		}

		// 循环浏览我们的用户列表，找到具有此id的用户。
		for (int count1 = 0; count1 < userList.size(); count1++) {
			tmpUser = (DushiUser) userList.elementAt(count1);

			if (tmpUser.id != userId)
				continue;

			// 就是这个。该用户以前在我们的聊天室吗？如果是，
			// 我们需要将它们从“当前发送到”列表中删除。
			String oldRoomName = tmpUser.chatroomName;
			if (oldRoomName.equals(mainPanel.currentRoom.name)) {
				if (!oldRoomName.equals(""))
					mainPanel.messagesArea.append("<<" + tmpUser.name + " "
							+ mainPanel.strings.get(thisClass, "movedto")
							+ " \"" + newRoomName + "\">>\n");

				// 从“当前发送到”列表中删除此名称。
				synchronized (list) {
					for (int count2 = 0; count2 < list.getItemCount(); count2++) {
						if (list.getItem(count2).equals(tmpUser.name)) {
							if (list.isIndexSelected(count2))
								list.select(0);
							list.remove(count2);
							break;
						}
					}
				}
			}
			// 用户是否正在进入我们的聊天室？
			else if (newRoomName.equals(mainPanel.currentRoom.name)) {
				if (!oldRoomName.equals(""))
					mainPanel.messagesArea.append("<<" + tmpUser.name + " "
							+ mainPanel.strings.get(thisClass, "entering")
							+ ">>\n");

				// 将此名称添加到我们的“当前发送到”列表中。
				synchronized (list) {
					list.add(tmpUser.name);
				}
			}

			// 设置用户的新聊天室名称。
			tmpUser.chatroomName = newRoomName;

			break;
		}

		// 如果有DushiRoomControlDialog，则获取房间列表更新。
		if (mainPanel.roomControlDialog != null)
			sendRoomList();
	}

	/**
	 * 告诉服务器我们要从当前聊天室引导用户。
	 */
	protected void sendBootUser(int userId, String roomName) {
		super.sendBootUser(mainPanel.clientId, roomName, userId);
	}

	/**
	 * 正在从当前聊天室启动此用户。
	 */
	private void receiveBootUser() throws IOException {
		DushiUser fromUser = null;
		String roomName = null;

		// 是谁？
		fromUser = readUser();

		roomName = iStream.readUTF();
		iStream.readInt(); // 丢出我们的用户id

		// 不要忽略这一点，即使启动我们的用户是被忽略的用户；
		// 被忽略的用户仍然有权从他们的聊天室引导我们，如果我
		// 们被引导，我们应该得到通知。

		if (fromUser == null)
			//确认。没有这样的用户。如果有人在适当的时候注销，就会发生这种情况。
			return;

		// 制作一个带有消息的对话框，以便用户注意到他们已经消失了。
		new DushiOkDialog(mainPanel.parentWindow, mainPanel.strings.get(
				thisClass, "booted"), mainPanel.strings.get(thisClass,
				"bootedfrom")
				+ " \""
				+ roomName
				+ "\" "
				+ mainPanel.strings.get(thisClass, "byuser")
				+ " "
				+ fromUser.name, mainPanel.strings.get("ok"), false)
				.setVisible(true);
	}

	/**
	 * 告诉服务器我们要禁止该用户进入当前聊天室。
	 */
	protected void sendBanUser(int userId, String roomName) {
		super.sendBanUser(mainPanel.clientId, roomName, userId);
	}

	/**
	 * 此用户已被禁止进入指定的聊天室。
	 */
	private void receiveBanUser() throws IOException {
		String roomName = null;

		roomName = iStream.readUTF();
		iStream.readInt(); // 丢出我们的用户id

		// 不要忽视这一点，即使禁止我们的用户是被忽视的用户；
		// 被忽视的用户仍然有权禁止我们进入他们的聊天室，如果
		// 我们被禁止，我们应该得到通知。

		// 制作一个带有消息的对话框，以便用户注意到他们已经消失了。
		new DushiOkDialog(mainPanel.parentWindow, mainPanel.strings.get(
				thisClass, "banned"), mainPanel.strings.get(thisClass,
				"bannedfrom") + " \"" + roomName + "\"",
				mainPanel.strings.get("ok"), false).setVisible(true);
	}

	/**
	 * 告诉服务器我们要取消禁止该用户进入当前聊天室。
	 */
	protected void sendAllowUser(int userId, String roomName) {
		super.sendAllowUser(mainPanel.clientId, roomName, userId);
	}

	/**
	 * 此用户已被取消禁止进入聊天室。
	 */
	private void receiveAllowUser() throws IOException {
		DushiUser fromUser = null;
		String roomName = null;

		// 是谁?
		fromUser = readUser();

		roomName = iStream.readUTF();
		iStream.readInt(); // 丢出我们的用户id

		if (fromUser == null)
			// 确认。没有这样的用户。如果有人在适当的时候注销，就会发生这种情况。
			return;

		if (isIgnoredUser(fromUser.name))
			return;

		// 制作一个包含消息的对话框，以便用户知道他们被允许返回。
		new DushiOkDialog(mainPanel.parentWindow, mainPanel.strings.get(
				thisClass, "allowed"), mainPanel.strings.get(thisClass,
				"allowedto") + " \"" + roomName + "\"",
				mainPanel.strings.get("ok"), false).setVisible(true);
	}

	/**
	 * 这告诉我们选择的收件人，我们正在打字或绘图。
	 */
	protected void sendActivity(short activity) {
		super.sendActivity(mainPanel.clientId, activity,
				mainPanel.getRecipientIds());
	}

	/**
	 * 一些用户正在进行一些活动，例如打字或绘图。
	 */
	private void receiveActivity() throws IOException {
		DushiUser fromUser = null;
		short activity = 0;
		int numForUsers = 0;

		//是谁?
		fromUser = readUser();

		// 阅读活动并放弃收件人列表（如果有的话）。
		activity = iStream.readShort();
		numForUsers = iStream.readInt();
		for (int count = 0; count < numForUsers; count++)
			iStream.readInt();

		if (fromUser == null)
			// 确认。没有这样的用户。如果有人在适当的时候注销，就会发生这种情况。
			return;

		// 我们是否忽略此用户？
		if (isIgnoredUser(fromUser.name))
			return;

		// 在活动窗口中设置适当的消息。
		String tmpString = "";
		if (activity == DushiCommand.ACTIVITY_DRAWING)
			tmpString = mainPanel.strings.get(thisClass, "drawing") + " "
					+ fromUser.name;
		else if (activity == DushiCommand.ACTIVITY_TYPING)
			tmpString = mainPanel.strings.get(thisClass, "typing") + " "
					+ fromUser.name;
		// 否则怎么办？忽略它。
		if (!mainPanel.activityField.getText().equals(tmpString))
			mainPanel.activityField.setText(tmpString);
	}

	/**
	 * 这将向选定的收件人发送一行聊天文字。
	 */
	protected void sendChatText(String data) {
		super.sendChatText(mainPanel.clientId,
				!mainPanel.sendToAllCheckbox.getState(), data,
				mainPanel.getRecipientIds());
	}

	/**
	 * 有传入的聊天文字。
	 */
	private void receiveChatText() throws IOException {
		DushiUser fromUser = null;
		boolean priv = false;
		String data = "";
		String output = "";
		int numForUsers = 0;

		// 这条消息是谁发来的？
		fromUser = readUser();

		// 这条消息是私有的吗？还要获取颜色值（未使用）、数据本身和收件人列表。
		priv = iStream.readBoolean();
		iStream.readShort(); // 目前未使用的颜色。
		data = iStream.readUTF();

		// 放弃收件人列表（如果有）。
		numForUsers = iStream.readInt();
		for (int count = 0; count < numForUsers; count++)
			iStream.readInt();

		if (fromUser != null) {
			if (isIgnoredUser(fromUser.name))
				return;

			// 在信息前面加上发送信息的用户的姓名。信息是公用的还是专用的？
			if (fromUser != null) {
				if (priv)
					output += "*"
							+ mainPanel.strings.get(thisClass, "privatefrom")
							+ " " + fromUser.name + "*> ";
				else
					output += fromUser.name + "> ";
			}
		}

		// 附加实际消息
		output += data;

		// 清除“活动”窗口
		mainPanel.activityField.setText("");

		// 将所有文本放入聊天窗口
		mainPanel.messagesArea.append(output);
	}

	/**
	 * 这将向选定的收件人发送一条图形线
	 */
	protected void sendLine(short colour, short startX, short startY,
			short endX, short endY, short thick) {
		super.sendLine(mainPanel.clientId, colour, startX, startY, endX, endY,
				thick, mainPanel.getRecipientIds());
	}

	/**
	 * 有一条传入的图形线。
	 */
	private void receiveLine() throws IOException {
		DushiUser fromUser = null;
		short colournum = 0;
		short startx = 0;
		short starty = 0;
		short endx = 0;
		short endy = 0;
		short thick = 0;
		int numForUsers = 0;
		Color colour;

		// 这条消息是谁发来的？
		fromUser = readUser();

		// 获取颜色值、x、y和厚度，以及收件人列表。
		colournum = iStream.readShort();
		startx = iStream.readShort();
		starty = iStream.readShort();
		endx = iStream.readShort();
		endy = iStream.readShort();
		thick = iStream.readShort();

		// 放弃收件人列表（如果有）。
		numForUsers = iStream.readInt();
		for (int count = 0; count < numForUsers; count++)
			iStream.readInt();

		if (fromUser != null)
			if (isIgnoredUser(fromUser.name))
				return;

		// 现在划线。
		colour = DushiCanvas.colourArray[colournum];
		mainPanel.canvas.drawLine(colour, startx, starty, endx, endy, thick,
				DushiCanvas.MODE_PAINT);
	}

	/**
	 * 这将向选定的收件人发送一个图形多边形。
	 */
	protected void sendPoly(short kind, short colour, short x, short y,
			short width, short height, short thick, boolean fill) {
		super.sendPoly(kind, mainPanel.clientId, colour, x, y, width, height,
				thick, fill, mainPanel.getRecipientIds());
	}

	/**
	 * 有一个传入的图形多边形。
	 */
	private void receivePoly(short kind) throws IOException {
		DushiUser fromUser = null;
		short colournum = 0;
		short x = 0;
		short y = 0;
		short height = 0;
		short width = 0;
		short thick = 0;
		boolean fill = false;
		int numForUsers = 0;
		Color colour;

		// 这条消息是谁发来的？
		fromUser = readUser();

		// 获取颜色值、x、y、宽度、高度、厚度和填充，以及收件人列表。
		colournum = iStream.readShort();
		x = iStream.readShort();
		y = iStream.readShort();
		width = iStream.readShort();
		height = iStream.readShort();
		thick = iStream.readShort();
		fill = iStream.readBoolean();

		// 放弃收件人列表（如果有）。
		numForUsers = iStream.readInt();
		for (int count = 0; count < numForUsers; count++)
			iStream.readInt();

		if (fromUser != null)
			if (isIgnoredUser(fromUser.name))
				return;

		// 获取颜色。
		colour = DushiCanvas.colourArray[colournum];

		// 现在绘制形状。
		if (kind == DushiCommand.RECT)
			mainPanel.canvas.drawRect(colour, x, y, width, height, fill, thick,
					DushiCanvas.MODE_PAINT);
		else if (kind == DushiCommand.OVAL)
			mainPanel.canvas.drawOval(colour, x, y, width, height, fill, thick,
					DushiCanvas.MODE_PAINT);
	}

	/**
	 * 这将向选定的收件人发送图形文本。
	 */
	protected void sendDrawText(short colour, short x, short y, short type,
			short attribs, short size, String text) {
		super.sendDrawText(mainPanel.clientId, colour, x, y, type, attribs,
				size, text, mainPanel.getRecipientIds());
	}

	/**
	 * 画布有一个传入的图形文本。
	 */
	private void receiveDrawText() throws IOException {
		DushiUser fromUser = null;
		short colournum = 0;
		short x = 0;
		short y = 0;
		short type = 0;
		short size = 0;
		short attribs = 0;
		String text = "";
		int numForUsers = 0;
		Color colour;

		// 这条消息是谁发来的？
		fromUser = readUser();

		// 获取颜色值、x、y、类型、大小和属性，以及收件人列表。
		colournum = iStream.readShort();
		x = iStream.readShort();
		y = iStream.readShort();
		type = iStream.readShort();
		attribs = iStream.readShort();
		size = iStream.readShort();
		text = iStream.readUTF();

		// 放弃收件人列表（如果有）。
		numForUsers = iStream.readInt();
		for (int count = 0; count < numForUsers; count++)
			iStream.readInt();

		if (fromUser != null)
			if (isIgnoredUser(fromUser.name))
				return;

		// 获取颜色
		colour = DushiCanvas.colourArray[colournum];

		// 现在绘制文本
		mainPanel.canvas.drawText(colour, x, y, type, attribs, size, text,
				DushiCanvas.MODE_PAINT);
	}

	/**
	 * 有一张传入的图片。
	 */
	private void receiveImage() throws IOException {
		int fromId = 0;
		DushiUser fromUser = null;
		String imageName = null;
		short xCoord = 0;
		short yCoord = 0;
		int totalSize = 0;
		int totalBlocks = 0;
		int blockSize = 0;
		byte[] data = null;
		int numRecipients = 0;
		DushiImageTransfer transfer = null;

		// 这个是谁的？
		fromId = iStream.readInt();
		fromUser = findUser(fromId);

		if (protocolVersion >= 2.2)
			// 读取图像名称
			imageName = iStream.readUTF();
		else {
			// 编一个。
			if (fromUser != null)
				imageName = (fromUser.name + Math.random());
			else
				imageName = ("unknown" + Math.random());
		}

		// 读取x和y坐标。
		xCoord = iStream.readShort();
		yCoord = iStream.readShort();

		if (protocolVersion >= 2.2) {
			// 图像数据以单独的IMAGEBLOCK命令发送。
			// 在这里，我们刚刚阅读了我们将收到的信息数量。
			totalSize = iStream.readInt();
			totalBlocks = iStream.readInt();
			blockSize = iStream.readInt();
		} else {
			// 读取实际图像数据。
			totalSize = iStream.readInt();
			totalBlocks = ((totalSize / blockSize) + (((totalSize % blockSize) > 0) ? 1
					: 0));
			data = new byte[totalSize];
			int read = 0;
			while (read < totalSize) {
				int tmp = iStream.read(data, read, (totalSize - read));
				if (tmp < 1)
					throw (new EOFException());
				read += tmp;
			}
		}

		// 放弃收件人列表（如果有）。
		numRecipients = iStream.readInt();
		for (int count = 0; count < numRecipients; count++)
			iStream.readInt();

		if (fromUser != null)
			if (isIgnoredUser(fromUser.name))
				return;

		transfer = new DushiImageTransfer(fromId, imageName, xCoord, yCoord,
				totalSize, totalBlocks, blockSize, null, null, mainPanel, null);

		if (protocolVersion >= 2.2)
			// 把它放在我们的图像传输列表中，等待所有数据到达。
			imageTransfers.put(imageName, transfer);
		else
			//我们已经掌握了所有的数据。
			transfer.putImageAll(data);
	}

	/**
	 * 向指定的用户发送“清除画布”信号。
	 */
	protected void sendClearCanv() {
		super.sendClearCanv(mainPanel.clientId, mainPanel.getRecipientIds());
	}

	/**
	 * 服务器告诉我们，另一个用户已经清除了画布。
	 */
	private void receiveClearCanv() throws IOException {
		DushiUser fromUser = null;
		int numForUsers = 0;

		// 是谁?
		fromUser = readUser();

		//放弃收件人列表（如果有）。
		numForUsers = iStream.readInt();
		for (int count = 0; count < numForUsers; count++)
			iStream.readInt();

		if (fromUser == null)
			// 确认。没有这样的用户。如果有人在适当的时候注销，就会发生这种情况。
			return;

		if (isIgnoredUser(fromUser.name))
			return;

		// 清除画布。
		mainPanel.canvas.clear();
		mainPanel.messagesArea.append("<<" + fromUser.name + " "
				+ mainPanel.strings.get(thisClass, "clearedcanvas") + ">>\n");
	}

	/**
	 * 向指定的用户发送寻呼信号。
	 */
	protected void sendPageUser() {
		super.sendPageUser(mainPanel.clientId, mainPanel.getRecipientIds());

		String[] selectedUsers;
		int numberUsers = 0;

		// 在我们自己的屏幕上打印一条消息，说明页面已发布。

		// 是否选中“全部发送”复选框？如果是这样，这意味着我们将发送给每个人。
		if (mainPanel.sendToAllCheckbox.getState())
			mainPanel.messagesArea.append("<<"
					+ mainPanel.strings.get(thisClass, "pagingall") + ">>\n");

		// 否则，请列出已被寻呼的用户。
		else {
			// 获取所选用户名的列表。
			selectedUsers = mainPanel.sendToList.getSelectedItems();

			// 有多少？
			numberUsers = selectedUsers.length;

			if (numberUsers == 1)
				mainPanel.messagesArea.append("<<"
						+ mainPanel.strings.get(thisClass, "paging") + " "
						+ selectedUsers[0] + ">>\n");
			else {
				mainPanel.messagesArea.append("<<"
						+ mainPanel.strings.get(thisClass, "pagingsome"));
				// 针对每项的循环。
				for (int count = 0; count < numberUsers; count++)
					mainPanel.messagesArea.append(" " + selectedUsers[count]);
				mainPanel.messagesArea.append(">>\n");
			}
		}
	}

	/**
	 * 有人在寻呼我们。
	 */
	private void receivePageUser() throws IOException {
		DushiUser fromUser = null;
		int numForUsers = 0;

		// 是谁？
		fromUser = readUser();

		// 放弃收件人列表（如果有）。
		numForUsers = iStream.readInt();
		for (int count = 0; count < numForUsers; count++)
			iStream.readInt();

		if (fromUser == null)
			// 确认。没有这样的用户。如果有人在适当的时候注销，就会发生这种情况。
			return;

		if (isIgnoredUser(fromUser.name))
			return;

		// 在会议区写一条消息，这样用户就可以看到自己被寻呼了。
		mainPanel.messagesArea.append("<<" + fromUser.name + " "
				+ mainPanel.strings.get(thisClass, "pagingyou") + ">>\n");

		// 如果用户没有关闭寻呼功能，请播放寻呼声音。
		// 否则，返回一条消息，说明页面已被拒绝。
		if (mainPanel.playSound) {
			// 如果这个Java版本低于1.2，我们就不能播放这个声音。
			double javaVersion = 0.0;
			try {
				javaVersion = new Double(System.getProperty("java.version")
						.substring(0, 3)).doubleValue();
			} catch (NumberFormatException e) {
			}

			if (javaVersion < 1.2)
				sendError(DushiCommand.ERROR_NOSOUND, fromUser.id);
			else {
				// 试着播放寻呼声音。
				try {
					URL soundURL = new URL(mainPanel.baseURL.getProtocol(),
							mainPanel.baseURL.getHost(),
							mainPanel.baseURL.getFile() + "DushiPage.au");
					Applet.newAudioClip(soundURL).play();
				} catch (Exception argh) {
					// 此客户端无法播放寻呼声音。

					// 试试哔哔声。
					for (int count = 0; count < 3; count++)
						mainPanel.getToolkit().beep();

					// 发送一条消息。
					sendError(DushiCommand.ERROR_NOSOUND, fromUser.id);
				}
			}
		} else
			// 用户不接受页面。发送消息。
			sendError(DushiCommand.ERROR_NOPAGE, fromUser.id);
	}

	/**
	 * 向指定用户的服务器发送即时消息。
	 */
	protected void sendInstantMess(int toId, String message) {
		super.sendInstantMess(mainPanel.clientId, toId, message);
	}

	/**
	 * 服务器正在向我们发送来自另一个用户的即时消息。
	 */
	private void receiveInstantMess() throws IOException {
		DushiUser fromUser = null;
		String message = "";

		// 是谁？
		fromUser = readUser();

		// 什么信息？
		iStream.readInt(); // 我们的id
		message = iStream.readUTF();

		if (fromUser == null)
			//确认。没有这样的用户。如果有人在适当的时候注销，就会发生这种情况
			return;

		if (isIgnoredUser(fromUser.name))
			return;

		// 是否存在现有的即时消息对话框？
		for (int count = 0; count < mainPanel.instantMessageDialogs.size(); count++) {
			DushiInstantMessageDialog tmp = (DushiInstantMessageDialog) mainPanel.instantMessageDialogs
					.elementAt(count);

			if (tmp.fromUser == fromUser) {
				// 是否存在现有的即时消息对话框？已经显示了一个对话框。只需向其发送附加消息。
				tmp.addMessage(message);
				return;
			}
		}

		// 将其作为对话框放在屏幕上
		mainPanel.instantMessageDialogs
				.addElement(new DushiInstantMessageDialog(mainPanel, fromUser,
						message));

		return;
	}

	/**
	 * 将我们的消息发送到指定用户的服务器。
	 */
	protected void sendLeaveMess(String toName, String message) {
		super.sendLeaveMess(mainPanel.clientId, toName, message);
	}

	/**
	 * 向服务器发送请求以检索我们存储的消息。
	 */
	protected void sendReadMess() {
		super.sendReadMess(mainPanel.clientId);
	}

	/**
	 * 服务器正在向我们发送所有存储的消息。
	 */
	private void receiveStoredMess() throws IOException {
		short numberMessages = 0;
		String from = "";
		String message = "";

		// 读取命令头
		numberMessages = iStream.readShort();

		if (numberMessages == 0) {
			new DushiOkDialog(mainPanel.parentWindow, mainPanel.strings.get(
					thisClass, "none"), mainPanel.strings.get(thisClass,
					"nounread"), mainPanel.strings.get("ok"), false)
					.setVisible(true);
			return;
		}

		// 消息的循环
		for (int count = 0; count < numberMessages; count++) {
			// 要传达什么信息？
			from = iStream.readUTF();
			message = iStream.readUTF();

			// 将其作为对话框放在屏幕上
			new DushiTextDialog(mainPanel.parentWindow, mainPanel.strings.get(
					thisClass, "messagefrom") + " " + from, message, 40, 10,
					TextArea.SCROLLBARS_VERTICAL_ONLY, false,
					mainPanel.strings.get("dismiss"));
		}

		// 告诉用户仅此而已
		new DushiOkDialog(mainPanel.parentWindow, mainPanel.strings.get(
				thisClass, "done"), mainPanel.strings.get(thisClass,
				"endmessages"), mainPanel.strings.get("ok"), false)
				.setVisible(true);
		return;
	}

	/**
	 * 将请求的错误通知发送给指定用户。
	 */
	protected void sendError(short code, int toId) {
		int[] recipient = new int[1];
		recipient[0] = toId;
		super.sendError(mainPanel.clientId, code, recipient);
	}

	/**
	 * 另一个用户向我们发送了错误代码。
	 */
	private void receiveError() throws IOException {
		DushiUser fromUser = null;
		short errorCode = 0;
		int numForUsers = 0;

		// 是谁？
		fromUser = readUser();

		// 错误代码是什么？
		errorCode = iStream.readShort();

		// 放弃收件人列表（如果有）。
		numForUsers = iStream.readInt();
		for (int count = 0; count < numForUsers; count++)
			iStream.readInt();

		if (fromUser == null)
			// 确认。没有这样的用户。如果有人在适当的时候注销，就会发生这种情况
			return;

		if (isIgnoredUser(fromUser.name))
			return;

		// 现在根据咨询类型做不同的事情
		switch (errorCode) {
		  case DushiCommand.ERROR_NOPAGE: {
			// 用户不接受我们的页面
			mainPanel.messagesArea.append("<<" + fromUser.name + " "
					+ mainPanel.strings.get(thisClass, "nopages") + ">>\n");
			break;
		  }
		  case DushiCommand.ERROR_NOSOUND: {
			// 此客户端无法播放声音
			mainPanel.messagesArea.append("<<" + fromUser.name + " "
					+ mainPanel.strings.get(thisClass, "nosound") + ">>\n");
			break;
		  }
		  default: {
			// 否则呢？
			mainPanel.messagesArea.append("<<"
					+ mainPanel.strings.get(thisClass, "unknownerror") + " "
					+ fromUser.name + ">>\n");
			break;
		  }
		}
	}

	/**
	 * 使用提供的参数和数据向服务器发送“发送电子邮件”请求。
	 */
	protected void sendSendEmail(String fromAddress, String toAddress,
			String subject, String text) {
		String mxHost = null;

		// 试着查找MX主机，这样服务器就不必为我们浪费时间了。
		try {
			mxHost = DushiEmail.mxLookup(toAddress);
		} catch (Exception e) {
		}
		if (mxHost == null)
			mxHost = "";

		super.sendSendEmail(mainPanel.clientId, mxHost, fromAddress, toAddress,
				subject, text);
	}

	/**
	 * 这将请求向所选用户发送文件。
	 */
	protected void sendFile(String fileName, long totalSize, int totalBlocks,
			int blockSize, int expireMins) {
		super.sendFile(mainPanel.clientId, fileName, totalSize, totalBlocks,
				blockSize, expireMins, mainPanel.getRecipientIds());
	}

	/**
	 * 有人想给我们寄一份文件。
	 */
	private void receiveFile() throws IOException {
		DushiUser fromUser = null;
		String fileName = null;
		long totalSize = 0;
		int totalBlocks = 0;
		int blockSize = 0;
		int expireMins = 0;
		int numForUsers = 0;

		// 这个是谁？
		fromUser = readUser();

		// 读取文件名、总大小、总块数和到期分钟数。
		fileName = iStream.readUTF();
		totalSize = iStream.readLong();
		totalBlocks = iStream.readInt();
		blockSize = iStream.readInt();
		expireMins = iStream.readInt();

		// 放弃收件人列表（如果有）
		numForUsers = iStream.readInt();
		for (int count = 0; count < numForUsers; count++)
			iStream.readInt();

		if (fromUser == null)
			// 确认。没有这样的用户。如果有人在适当的时候注销，就会发生这种情况
			return;

		if (isIgnoredUser(fromUser.name))
			return;

		// 设置文件传输，等待客户接受。
		fileTransfers.put(fileName, new DushiClientFileTransfer(mainPanel,
				fromUser, fileName, totalSize, totalBlocks, blockSize,
				expireMins));
	}

	/**
	 * 告诉服务器我们是否要接受由我们收到的“发送文件”请求启动的文件传输。
	 */
	protected void sendFileAccept(int toId, String fileName, boolean accept) {
		super.sendFileAccept(mainPanel.clientId, toId, fileName, accept);
	}

	/**
	 * 服务器告诉我们有人接受或拒绝了我们启动文件传输的请求。
	 */
	private void receiveFileAccept() throws IOException {
		DushiUser fromUser = null;
		String fileName = null;
		boolean accepted = false;
		DushiClientFileTransfer transfer = null;

		// 这个是谁？
		fromUser = readUser();

		// 丢弃我们的id，读取文件名，以及是否已接受传输。
		iStream.readInt(); // 我们的id。
		fileName = iStream.readUTF();
		accepted = iStream.readBoolean();

		if (fromUser == null)
			// 确认。没有这样的用户。如果有人在适当的时候注销，就会发生这种情况。
			return;

		transfer = (DushiClientFileTransfer) fileTransfers.get(fileName);
		if (transfer != null)
			transfer.notifyAccepted(fromUser, accepted);
	}

	/**
	 * 通过“发送文件”请求启动的文件传输，向服务器发送一个文件块。
	 */
	protected void sendFileBlock(String fileName, int blockNumber, byte[] data) {
		super.sendFileBlock(mainPanel.clientId, fileName, blockNumber, data);
	}

	/**
	 * 服务器正在发送一个文件块。找到由“发送文件”请求启动的文件传输，
	 * 读取块，并将其添加到文件传输中。
	 */
	private void receiveFileBlock() throws IOException {
		DushiUser fromUser = null;
		String fileName = null;
		int blockNumber = 0;
		int dataLength = 0;
		byte[] data = null;
		DushiClientFileTransfer transfer = null;

		// 这个是谁？
		fromUser = readUser();

		// 读取文件名、块号和数据长度。
		fileName = iStream.readUTF();
		blockNumber = iStream.readInt();
		dataLength = iStream.readInt();

		// 现在读取数据。
		data = new byte[dataLength];
		int read = 0;
		while (read < dataLength) {
			int tmp = iStream.read(data, read, (dataLength - read));
			if (tmp < 1)
				throw (new EOFException());
			read += tmp;
		}

		transfer = (DushiClientFileTransfer) fileTransfers.get(fileName);
		if ((transfer != null) && (transfer.fromUser == fromUser)
				&& transfer.accepted)
			transfer.putFileBlock(blockNumber, data);
	}

	/**
	 * 服务器告诉我们，我们参与的文件传输已被取消。
	 */
	private void receiveFileCancel() throws IOException {
		int fromUserId = 0;
		String fileName = null;
		DushiClientFileTransfer transfer = null;

		// 这个是谁？
		fromUserId = iStream.readInt();

		// 读取文件名
		fileName = iStream.readUTF();

		// 查找文件传输
		transfer = (DushiClientFileTransfer) fileTransfers.get(fileName);
		if ((transfer != null) && (transfer.fromUser.id == fromUserId)) {
			transfer.cancel();
			fileTransfers.remove(fileName);
		}
	}

	/**
	 * 服务器正在发送一个图像块。找到由“图像”请求启动的图像传输，
	 * 读取块，并将其添加到图像传输中。
	 */
	private void receiveImageBlock() throws IOException {
		DushiUser fromUser = null;
		String imageName = null;
		int blockNumber = 0;
		int dataLength = 0;
		byte[] data = null;
		DushiImageTransfer transfer = null;

		// 这个是谁？
		fromUser = readUser();

		// 读取图像名称、块号和数据长度。
		imageName = iStream.readUTF();
		blockNumber = iStream.readInt();
		dataLength = iStream.readInt();

		// 现在读取数据。
		data = new byte[dataLength];
		int read = 0;
		while (read < dataLength) {
			int tmp = iStream.read(data, read, (dataLength - read));
			if (tmp < 1)
				throw (new EOFException());
			read += tmp;
		}

		if (fromUser != null)
			if (isIgnoredUser(fromUser.name))
				return;

		transfer = (DushiImageTransfer) imageTransfers.get(imageName);
		if (transfer != null)
			transfer.putImageBlock(blockNumber, data);

		// 如果我们已经收到了所有内容，请将其添加到画布中。
		if (transfer.gotAll)
			imageTransfers.remove(transfer.imageName);
	}

	protected void lostConnection() {
		if (!stop) {
			shutdown(true);
			mainPanel.offline();
		}
	}

	protected synchronized void shutdown(boolean notifyUser) {
		if (!online)
			return;

		super.shutdown();

		// 清空我们的用户列表
		userList.removeAllElements();

		// 不再有聊天室
		mainPanel.currentRoom = null;

		// 去掉任何可能挂在周围的“聊天室列表”、“聊天室控制”或即时消息对话框。
		if (mainPanel.roomsDialog != null) {
			mainPanel.roomsDialog.dispose();
			mainPanel.roomsDialog = null;
		}
		if (mainPanel.roomControlDialog != null) {
			mainPanel.roomControlDialog.dispose();
			mainPanel.roomControlDialog = null;
		}
		for (int count = 0; count < mainPanel.instantMessageDialogs.size(); count++) {
			DushiInstantMessageDialog tmp = (DushiInstantMessageDialog) mainPanel.instantMessageDialogs
					.elementAt(count);
			tmp.dispose();
		}
		mainPanel.instantMessageDialogs.removeAllElements();

		if (notifyUser)
			// 向用户发送消息
			new DushiOkDialog(mainPanel.parentWindow, mainPanel.strings.get(
					thisClass, "disconnected"), mainPanel.strings.get(
					thisClass, "disconnectedfrom") + " " + mainPanel.host,
					mainPanel.strings.get("ok"), true).setVisible(true);

		return;
	}
}
